//
// Created by #Suyghur, on 2021/06/15.
//

#include <third_part/openssl/ossl_typ.h>
#include <third_part/openssl/bio.h>
#include <third_part/openssl/pem.h>
#include <logger.h>
#include <toolkit.h>
#include "include/base64_utils.h"
#include "include/rsa/rsa_utils.h"

/**
 * 公钥加密
 */
std::string RsaUtils::EncryptByPublicKey(std::string public_key, std::string src) {

    std::string key = GeneratePublicKey(public_key);
    std::string raw = Encrypt(key, src);
    if (raw.empty()) {
        return nullptr;
    }

//    LOGD("raw : %s", raw.c_str());

    std::string base64_raw = Base64Utils::Encode(raw);
    if (base64_raw.empty()) {
        return nullptr;
    }

    return base64_raw;
}

/**
 * 公钥解密
 */
std::string RsaUtils::DecryptByPublicKey(std::string public_key, std::string raw) {
    std::string key = GeneratePublicKey(public_key);
    LOGD("decrypt pubkey : %s", key.c_str());

    LOGD("decrypt key : %s", raw.c_str());
    return Decrypt(key, raw);
}

std::string RsaUtils::GeneratePublicKey(std::string encode_key) {
    std::string public_key = encode_key;
    size_t base64_length = 64;//每64个字符一行
    size_t public_key_length = encode_key.size();
    for (size_t i = base64_length; i < public_key_length; i += base64_length) {
        //每base64Length个字符，增加一个换行
        if (encode_key[i] != '\n') {
            public_key.insert(i, "\n");
        }
        i++;
    }
    //最前面追加公钥begin字符串
    public_key.insert(0, "-----BEGIN PUBLIC KEY-----\n");
    //最前面追加公钥end字符串
    public_key.append("\n-----END PUBLIC KEY-----");
    return public_key;
}

/**
 * 公钥加密
 */
std::string RsaUtils::Encrypt(const std::string &public_key, const std::string &src) {
    BIO *bio = nullptr;
    RSA *rsa_public_key = nullptr;

    //从字符串读取RSA公钥串
    if ((bio = BIO_new_mem_buf((void *) public_key.c_str(), -1)) == nullptr) {
        return "";
    }
    //读取公钥
    rsa_public_key = PEM_read_bio_RSA_PUBKEY(bio, nullptr, nullptr, nullptr);

    //异常处理
    if (rsa_public_key == nullptr) {
        //资源释放
        BIO_free_all(bio);
        //清除管理CRYPTO_EX_DATA的全局hash表中的数据，避免内存泄漏
        CRYPTO_cleanup_all_ex_data();
        return "";
    }

    //rsa模的位数
    int rsa_size = RSA_size(rsa_public_key);

    //RSA_PKCS1_PADDING 最大加密长度 为 128 -11
    //RSA_NO_PADDING 最大加密长度为  128
    //rsa_size = rsa_size - RSA_PKCS1_PADDING_SIZE;

    //动态分配内存，用于存储加密后的密文
    auto *raw = static_cast<unsigned char *>(malloc(rsa_size + 1));
    //填充0
    memset(raw, 0, rsa_size + 1);

    //明文长度
    int len = src.length();

    //加密，返回值为加密后的密文长度，-1表示失败
    int status = RSA_public_encrypt(len, (const unsigned char *) src.c_str(), raw, rsa_public_key, RSA_PKCS1_PADDING);
    //异常处理
    if (status < 0) {
        //资源释放
        free(raw);
        BIO_free_all(bio);
        RSA_free(rsa_public_key);
        //清除管理CRYPTO_EX_DATA的全局hash表中的数据，避免内存泄漏
        CRYPTO_cleanup_all_ex_data();
        return "";
    }
    //赋值密文
    std::string result((char *) raw, status);
    //资源释放
    free(raw);
    BIO_free_all(bio);
    RSA_free(rsa_public_key);
    //清除管理CRYPTO_EX_DATA的全局hash表中的数据，避免内存泄漏
    CRYPTO_cleanup_all_ex_data();
    return result;
}

/**
 * 公钥解密
 */
std::string RsaUtils::Decrypt(const std::string &public_key, const std::string &raw) {
    BIO *bio = nullptr;
    RSA *rsa_public_key = nullptr;

    //从字符串读取RSA公钥串
    if ((bio = BIO_new_mem_buf((void *) public_key.c_str(), -1)) == nullptr) {
        return "";
    }
    //读取公钥
    rsa_public_key = PEM_read_bio_RSA_PUBKEY(bio, nullptr, nullptr, nullptr);

    //异常处理
    if (rsa_public_key == nullptr) {
        //资源释放
        BIO_free_all(bio);
        //清除管理CRYPTO_EX_DATA的全局hash表中的数据，避免内存泄漏
        CRYPTO_cleanup_all_ex_data();
        return "";
    }

    //rsa模的位数
    int rsa_size = RSA_size(rsa_public_key);

    //动态分配内存，用于存储加密后的密文
    auto *src = static_cast<unsigned char *>(malloc(rsa_size + 1));
    //填充0
    memset(src, 0, rsa_size + 1);
    //加密，返回值为加密后的密文长度，-1表示失败
    auto *data = reinterpret_cast<unsigned char *>(Base64Utils::Decode(raw.c_str()));
    int len = strlen((char *) data);
    std::string d = (char *) data;

    std::string data_encode = Base64Utils::Encode((char *) data);

    int status = RSA_public_decrypt(len, data, src, rsa_public_key, RSA_PKCS1_PADDING);
    //解密
    //int ret = RSA_public_decrypt(clearText.length(), (const unsigned char*)clearText.c_str(), (unsigned char*)encryptedText, rsa, RSA_PKCS1_PADDING);
    //异常处理
    if (status < 0) {
        LOGD("status : %d", status);

        //资源释放
        free(src);
        free(data);
        BIO_free_all(bio);
        RSA_free(rsa_public_key);
        //清除管理CRYPTO_EX_DATA的全局hash表中的数据，避免内存泄漏
        CRYPTO_cleanup_all_ex_data();
        return "";
    }

    //赋值密文
    std::string result((char *) src, status);

    //资源释放
    free(data);
    free(src);
    BIO_free_all(bio);
    RSA_free(rsa_public_key);
    //清除管理CRYPTO_EX_DATA的全局hash表中的数据，避免内存泄漏
    CRYPTO_cleanup_all_ex_data();
    LOGD("result : %s", result.c_str());
    return result;

}

std::string RsaUtils::DecryptByPublicKey2(JNIEnv *env, std::string public_key, std::string raw) {

//    std::string key = GeneratePublicKey(public_key);
//    LOGD("decrypt pubkey : %s", key.c_str());
//
//    LOGD("decrypt key : %s", raw.c_str());
    return Decrypt2(env, public_key, raw);
}

std::string RsaUtils::Decrypt2(JNIEnv *env, const std::string &public_key, const std::string &enc) {
    jclass clz = env->FindClass("cn/yyxx/support/encryption/rsa/RsaUtils");
    if (clz == nullptr) {
        LOGE("rsa impl clz is nullptr !!!");
        return "";
    }

    const char *method_name = "decryptByPublicKey";
    const char *sig = "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;";
    jmethodID mid = env->GetStaticMethodID(clz, method_name, sig);
    jstring key_ = env->NewStringUTF(public_key.c_str());
    jstring enc_ = env->NewStringUTF(enc.c_str());
    auto result = (jstring) env->CallStaticObjectMethod(clz, mid, key_, key_);
    env->DeleteLocalRef(clz);
    env->DeleteLocalRef(key_);
    env->DeleteLocalRef(enc_);
//    free((void *) method_name);
//    free((void *) sig);
    return ToolKit::JString2String(env, result);
}



